"""Common code for metadata"""

from abc import ABCMeta, abstractmethod


class Descriptor(metaclass=ABCMeta):
    """The descriptor base class"""

    @property
    @abstractmethod
    def descriptor_type(self) -> str:
        """The descriptor type

        Returns:
            str: The type of the descriptor
        """



"""Meta data utilities"""

import inspect
import sys
from types import ModuleType
from typing import Any, Optional, Type


def make_file_relative(file: Optional[str]) -> Optional[str]:
    """Make a file path relative

    Args:
        file (Optional[str]): The file path

    Returns:
        Optional[str]: The relative file path
    """
    if file is None:
        return None

    for path in sys.path:
        if file.startswith(path):
            return file[len(path)+1:]

    return file


def is_named_tuple_type(obj: Type) -> bool:
    """Check if a type is a named tuple

    Args:
        obj (Type): The type to check

    Returns:
        bool: True if the type is a named tuple.
    """
    if tuple not in getattr(obj, '__bases__', []):
        return False

    fields = getattr(obj, '_fields', None)
    if not isinstance(fields, tuple):
        return False

    return all(isinstance(field, str) for field in fields)


def is_child_module(parent: ModuleType, child: Any) -> bool:
    """Check if a module is a child of another module

    Args:
        parent ([type]): The parent module
        child ([type]): The child to test

    Returns:
        bool: True if the child is a child module of the parent
    """

    return (
        inspect.ismodule(child) and
        getattr(child, '__package__', '').startswith(
            getattr(parent, '__package__', '')
        )
    )



"""A sample extension"""

import inspect
import re
from typing import Any, List, Optional
import xml.etree.ElementTree as etree
from xml.etree.cElementTree import Element

from jinja2 import (
    Environment,
    BaseLoader,
    PackageLoader,
    FileSystemLoader,
    select_autoescape
)
from markdown.blockparser import BlockParser
from markdown.blockprocessors import BlockProcessor


from .metadata import (
    Descriptor,
    ModuleDescriptor,
    CallableDescriptor,
    ClassDescriptor
)
from .utils import import_from_string


class AutodocBlockProcessor(BlockProcessor):
    """An inline processor for Python documentation"""

    def __init__(
            self,
            parser: BlockParser,
            *,
            class_from_init: bool = True,
            ignore_dunder: bool = True,
            ignore_private: bool = True,
            ignore_all: bool = False,
            ignore_inherited: bool = True,
            prefer_docstring: bool = True,
            follow_module_tree: bool = False,
            template_folder: Optional[str] = None,
            template_file: str = "main.jinja2"
    ) -> None:
        """An inline processor for **Python** documentation

        Args:
            pattern ([type]): The regular expression to match
            md (BlockParser): The parser.
            class_from_init (bool, optional): If True use the docstring from
                the <span>&#95;&#95;</span>init<span>&#95;&#95;</span> function
                for classes. Defaults to True.
            ignore_dunder (bool, optional): If True ignore
                <span>&#95;&#95;</span>XXX<span>&#95;&#95;</span> functions.
                Defaults to True.
            ignore_private (bool, optional): If True ignore private methods
                (those prefixed <span>&#95;</span>XXX). Defaults to True.
            ignore_all (bool): If True ignore the
                <span>&#95;&#95;</span>all<span>&#95;&#95;</span> member.
            ignore_inherited (bool): If True ignore inherited members.
            prefer_docstring (bool): If true prefer the docstring.
            follow_module_tree (bool): If true follow the module tree.
            template_folder (Optional[str], optional): The template folder,
                Defaults to None.
            template_file (Optional[str], optional): The template file to use,
                Defaults to "main.jinja2".
        """
        super().__init__(parser)
        self.class_from_init = class_from_init
        self.ignore_dunder = ignore_dunder
        self.ignore_private = ignore_private
        self.ignore_all = ignore_all
        self.ignore_inherited = ignore_inherited
        self.follow_module_tree = follow_module_tree
        self.prefer_docstring = prefer_docstring
        if template_folder:
            loader: BaseLoader = FileSystemLoader(template_folder)
        else:
            loader = PackageLoader('jetblack_markdown', 'templates')
        self.env = Environment(
            loader=loader,
            autoescape=select_autoescape(['html', 'xml'])
        )
        self.env.filters['md_format'] = self._md_format
        self.template = self.env.get_template(template_file)
        self._pattern = re.compile(r'@\[([^\]]+)\]')
        self._match: Optional[re.Match[str]] = None

    def test(self, parent: Element, block: str) -> bool:
        self._match = self._pattern.match(block)
        return self._match is not None

    def run(self, parent: Element, blocks: List[str]) -> Optional[bool]:

        assert self._match is not None
        import_str = self._match.group(1)
        if not import_str:
            return False

        html_text = self._render(import_str)
        element = etree.fromstring(html_text)
        parent.append(element)

        blocks.pop(0)

        return True

    def _render(self, import_str: str) -> str:
        obj = import_from_string(import_str)
        descriptor = self._make_descriptor(obj)
        html_text = self.template.render(
            obj=descriptor
        )
        return html_text

    def _make_descriptor(self, obj: Any) -> Descriptor:
        if inspect.ismodule(obj):
            return ModuleDescriptor.create(
                obj,
                self.class_from_init,
                self.ignore_dunder,
                self.ignore_private,
                self.ignore_all,
                self.ignore_inherited,
                self.prefer_docstring,
                self.follow_module_tree
            )
        elif inspect.isclass(obj):
            return ClassDescriptor.create(
                obj,
                self.class_from_init,
                self.ignore_dunder,
                self.ignore_private,
                self.ignore_inherited,
                prefer_docstring=self.prefer_docstring
            )
        elif inspect.isfunction(obj):
            return CallableDescriptor.create(
                obj,
                prefer_docstring=self.prefer_docstring
            )
        else:
            raise RuntimeError("Unhandled descriptor")

    def _md_format(self, text: str) -> str:
        parent = Element("div")
        self.parser.parseChunk(parent, text)
        children = next(iter(parent))
        buf = (
            etree.tostring(children[0])
            if len(children) == 1
            else etree.tostring(parent)
        )
        return buf.decode('utf-8')



"""A test placeholder"""

import xml.etree.ElementTree as etree

import markdown

from jetblack_markdown.autodoc import AutodocExtension
from jetblack_markdown.latex2mathml import Latex2MathMLExtension


def test_smoketest():
    content = """
An inline formula looks like: $x=\frac{-b\pm\sqrt{b^2-4ac} }{2a}$.

A block looks like:

$$
x=\frac{-b\pm\sqrt{b^2-4ac} }{2a}
$$

The outer `<math>` tag has the HTML class `"latex2mathml"`.

Here is some API documentation.

@[jetblack_markdown.latex2mathml]
"""
    extensions = [
        AutodocExtension(class_from_init=True),
        Latex2MathMLExtension()
    ]
    output = markdown.markdown(content, extensions=extensions)
    assert output is not None


def test_etree():
    tree = etree.fromstring('<div>Hello</div>')
    print(tree)



"""A markdown extension for creating documentation"""

from markdown import Markdown
from markdown.extensions import Extension

from .autodoc_processor import AutodocBlockProcessor

_DOCSTRING_RE = r'@\[([^\]]+)\]'

__all__ = [
    "AutodocExtension",
    "makeExtension"
]


class AutodocExtension(Extension):
    """The autodoc extension.

    Reference as "jetblack_markdown.autodoc"
    """

    def __init__(self, *args, **kwargs) -> None:
        self.config = {
            'class_from_init': [False, 'Class documentation is on __init__'],
            'ignore_dunder': [True, 'Ignore dunder methods'],
            'ignore_private': [True, 'Ignore private methods'],
            'ignore_all': [False, 'Ignore the __all__ member'],
            'ignore_inherited': [True, 'Ignore inherited members'],
            'prefer_docstring': [True, 'Prefer the docstring'],
            'follow_module_tree': [False, 'Follow the module tree'],
            'template_folder': ['', 'The template folder'],
            'template_file': ['main.jinja2', 'The template file to use'],
        }
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md: Markdown) -> None:
        class_from_init = self.getConfig('class_from_init')
        ignore_dunder = self.getConfig('ignore_dunder')
        ignore_private = self.getConfig('ignore_private')
        ignore_all = self.getConfig('ignore_all')
        ignore_inherited = self.getConfig('ignore_inherited')
        prefer_docstring = self.getConfig('prefer_docstring')
        follow_module_tree = self.getConfig('follow_module_tree')
        template_folder = self.getConfig('template_folder')
        template_file = self.getConfig('template_file')
        md.parser.blockprocessors.register(
            AutodocBlockProcessor(
                md.parser,
                class_from_init=class_from_init,
                ignore_dunder=ignore_dunder,
                ignore_private=ignore_private,
                ignore_all=ignore_all,
                ignore_inherited=ignore_inherited,
                prefer_docstring=prefer_docstring,
                follow_module_tree=follow_module_tree,
                template_folder=template_folder,
                template_file=template_file
            ),
            'autodoc',
            200
        )


# pylint: disable=invalid-name
def makeExtension(*args, **kwargs) -> Extension:
    """Make the extension

    This hook *function* gets picked up by the markdown processor when the
    extension is listed

    ```python
    output = markdown.markdown(
        content, extensions=[
            "admonition",
            "codehilite",
            "jetblack_markdown.autodoc",
        ])
    print(output)
    ```

    Args:
        *args: Positional arguments from the markdown processor
        **kwargs: Keyword arguments from the markdown processor

    Returns:
        Extension: The markdown extension
    """
    return AutodocExtension(*args, **kwargs)



"""A Latex to MathML markdown processor"""

from functools import partial
import re
from typing import (
    List,
    Optional,
    Tuple,
    Union
)
import xml.etree.ElementTree as etree
from xml.etree.cElementTree import Element

from markdown import Markdown
from markdown.inlinepatterns import InlineProcessor
from markdown.blockparser import BlockParser
from markdown.blockprocessors import BlockProcessor

from latex2mathml.converter import convert_to_element

HTML_CLASS = "latex2mathml"


class Latex2MathMLInlineProcessor(InlineProcessor):
    """An inline processor for converting Latex to MathML"""

    def __init__(
            self,
            pattern,
            md: Optional[Markdown] = None,
    ) -> None:
        super().__init__(pattern, md=md)

    def handleMatch(
            self,
            matches: re.Match,
            data: str
    ) -> Tuple[Optional[Union[etree.Element, str]], Optional[int], Optional[int]]:
        latex = matches.group(1)
        if not latex:
            return None, None, None

        element = convert_to_element(latex.strip())
        element.set("class", HTML_CLASS)
        del element.attrib['xmlns']

        start = matches.start(0)
        end = matches.end(0)
        return element, start, end


class Latex2MathMLBlockProcessor(BlockProcessor):
    """An block processor for converting Latex to MathML"""

    def __init__(self, parser: BlockParser):
        super().__init__(parser)
        self._pattern = re.compile(
            r' *\$\$\n(.*)\n\$\$ *'
        )
        self._match: Optional[re.Match[str]] = None

    def test(self, parent: Element, block: str) -> bool:
        self._match = self._pattern.match(block)
        return self._match is not None

    def run(self, parent: Element, blocks: List[str]) -> Optional[bool]:

        assert self._match is not None
        latex = self._match.group(1)
        if not latex:
            return False

        element = convert_to_element(
            latex.strip(),
            display='block',
            parent=parent
        )
        element.set("class", HTML_CLASS)
        del element.attrib['xmlns']

        blocks.pop(0)

        return True



"""JetBlack Markdown"""

from .autodoc import AutodocExtension, AutodocBlockProcessor
from .latex2mathml import Latex2MathMLExtension, Latex2MathMLInlineProcessor

__all__ = [
    'AutodocExtension',
    'AutodocBlockProcessor',
    'Latex2MathMLExtension',
    'Latex2MathMLInlineProcessor'
]



r"""A Latex to MathML markdown extension.

Documentation can be written inline like this: $ x=\frac{-b\pm\sqrt{b^2-4ac} }{2a} $.

Or in a block style:

$$
x=\frac{-b\pm\sqrt{b^2-4ac} }{2a}
$$
"""

from markdown import Markdown
from markdown.extensions import Extension

from .latex2mathml_processor import Latex2MathMLInlineProcessor, Latex2MathMLBlockProcessor


__all__ = [
    "Latex2MathMLExtension",
    "makeExtension"
]


class Latex2MathMLExtension(Extension):

    RE = r'\$([^$\n]+)\$'

    def __init__(self, *args, **kwargs) -> None:
        self.config = {}
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md: Markdown) -> None:
        md.inlinePatterns.register(
            Latex2MathMLInlineProcessor(
                self.RE,
                md,
            ),
            'mathml',
            50
        )
        md.parser.blockprocessors.register(
            Latex2MathMLBlockProcessor(md.parser),
            'mathml',
            50
        )


# pylint: disable=invalid-name
def makeExtension(*args, **kwargs) -> Extension:
    """Make the extension

    This hook *function* gets picked up by the markdown processor when the
    extension is listed

    ```python
    output = markdown.markdown(
        content, extensions=[
            "admonition",
            "codehilite",
            "jetblack_markdown.latex2mathml",
        ])
    print(output)
    ```

    Args:
        *args: Positional arguments from the markdown processor
        **kwargs: Keyword arguments from the markdown processor

    Returns:
        Extension: The markdown extension
    """
    return Latex2MathMLExtension(*args, **kwargs)



"""Classes"""

from __future__ import annotations
import inspect
from typing import (
    Any,
    Dict,
    List,
    Optional
)

import docstring_parser

from .arguments import ArgumentDescriptor
from .callables import CallableDescriptor, CallableType
from .common import Descriptor
from .properties import PropertyDescriptor
from .utils import make_file_relative, is_named_tuple_type


def _get_docstring(
        obj: Any,
        members: Dict[str, Any],
        class_from_init: bool,
        is_named_tuple: bool
) -> docstring_parser.Docstring:
    docstring = docstring_parser.parse(
        inspect.getdoc(
            members.get('__init__', obj)
            if class_from_init and not is_named_tuple else obj
        ) or ''
    )

    # There has to be a better way to detect an empty __init__.
    if docstring.short_description == 'Initialize self.  See help(type(self)) for accurate signature.':
        docstring = docstring_parser.parse(inspect.getdoc(obj) or '')

    return docstring


class ClassDescriptor(Descriptor):
    """A class descriptor"""

    def __init__(
            self,
            name: str,
            summary: Optional[str],
            description: Optional[str],
            constructor: Optional[CallableDescriptor],
            attributes: List[ArgumentDescriptor],
            properties: List[PropertyDescriptor],
            class_methods: List[CallableDescriptor],
            methods: List[CallableDescriptor],
            examples: Optional[List[str]],
            module: str,
            package: Optional[str],
            file: Optional[str],
            bases: List[ClassDescriptor]
    ) -> None:
        """A class descriptor

        Args:
            name (str): The class name
            summary (Optional[str]): The docstring summary
            description (Optional[str]): The docstring description
            constructor (Optional[CallableDescriptor]): The constructor
            attributes (List[ArgumentDescriptor]): The class attributes
            properties (List[PropertyDescriptor]): The class properties
            class_methods (List[CallableDescriptor]): The class methods
            methods (List[CallableDescriptor]): The class methods
            examples (Optional[List[str]]): Examples from the docstring
            module (str): The module
            package (Optional[str]): The package
            file (Optional[str]): The file,
            bases (List[ClassDescription]): The base classes
        """
        self.name = name
        self.summary = summary
        self.description = description
        self.constructor = constructor
        self.attributes = attributes
        self.properties = properties
        self.class_methods = class_methods
        self.methods = methods
        self.examples = examples
        self.module = module
        self.package = package
        self.file = file
        self.bases = bases

    @property
    def descriptor_type(self) -> str:
        return "class"

    def __repr__(self) -> str:
        return f'{self.name} - {self.summary}'

    @classmethod
    def create(
            cls,
            obj: Any,
            class_from_init: bool,
            ignore_dunder: bool,
            ignore_private: bool,
            ignore_inherited: bool,
            importing_module: Optional[str] = None,
            prefer_docstring: bool = True,
            imported_from_all: bool = False
    ) -> ClassDescriptor:
        """Create a class

        Args:
            obj (Any): The class
            class_from_init (bool): If True take the docstring from the init function
            ignore_dunder (bool): If True ignore
                <span>&#95;&#95;</span>XXX<span>&#95;&#95;</span> functions
            ignore_private (bool): If True ignore private methods (those
                prefixed <span>&#95;</span>XXX)
            ignore_inherited (bool): If True ignore inherited methods
            importing_module (Optional[str], optional): The importing module, defaults to None
            prefer_docstring (bool): If true prefer the docstring.
            imported_from_all (bool): If true the class if defined in the `__init__.py`.

        Returns:
            ClassDescriptor: The class descriptor
        """
        is_named_tuple = is_named_tuple_type(obj)

        valid_names: List[str] = []
        valid_names.extend(getattr(obj, '__dict__', {}).keys())
        valid_names.extend(getattr(obj, '__slots__', []))

        members: Dict[str, Any] = {
            name: value
            for name, value in inspect.getmembers(obj)
            if not ignore_inherited or name in valid_names
        }

        docstring = _get_docstring(
            obj, members,
            class_from_init,
            is_named_tuple
        )
        name = obj.__qualname__ if hasattr(
            obj,
            '__qualname__'
        ) else obj.__name__
        summary = docstring.short_description if docstring else None
        description = docstring.long_description if docstring else None

        attributes: List[ArgumentDescriptor] = []
        if docstring:
            attrs = [
                (meta.args[1], meta.description)
                for meta in docstring.meta
                if 'attribute' in meta.args
            ]
            for attr_details, attr_desc in attrs:
                attr_name, _sep, attr_type = attr_details.partition(' ')
                attr_type = attr_type.strip('()')
                attributes.append(
                    ArgumentDescriptor(attr_name, attr_type, attr_desc)
                )
        properties: List[PropertyDescriptor] = []
        methods: List[CallableDescriptor] = []
        class_methods: List[CallableDescriptor] = []
        for member_name, member in members.items():
            if member_name == '__init__' or (
                    ignore_dunder and
                    member_name.startswith('__') and
                    member_name.endswith('__')
            ) or (ignore_private and member_name.startswith('_')):
                continue

            if member.__class__ is property:
                properties.append(
                    PropertyDescriptor.create(
                        member,
                        obj,
                        member_name
                    )
                )
            elif inspect.isfunction(member):
                # Instance methods
                methods.append(
                    CallableDescriptor.create(
                        member,
                        callable_type=CallableType.METHOD,
                        prefer_docstring=prefer_docstring,
                        qualifier=name
                    )
                )
            elif inspect.ismethod(member):
                # Class methods
                class_methods.append(
                    CallableDescriptor.create(
                        member,
                        callable_type=CallableType.CLASS_METHOD,
                        prefer_docstring=prefer_docstring,
                        qualifier=name
                    )
                )

        examples: Optional[List[str]] = [
            meta.description or ''
            for meta in docstring.meta
            if 'examples' in meta.args
        ] if docstring is not None else None

        module_obj = inspect.getmodule(obj)
        module = importing_module or obj.__module__
        package = module_obj.__package__ if module_obj else None
        file = make_file_relative(
            module_obj.__file__
            if module_obj and hasattr(module_obj, '__file__')
            else None
        )

        bases = [
            ClassDescriptor.create(
                base,
                class_from_init=class_from_init,
                ignore_dunder=ignore_dunder,
                ignore_private=ignore_private,
                ignore_inherited=ignore_inherited,
                importing_module=importing_module,
                prefer_docstring=prefer_docstring
            )
            for base in getattr(obj, '__bases__', [])
            if base is not object
        ]

        try:
            signature = inspect.signature(obj)
            constructor = CallableDescriptor.create(
                obj,
                signature,
                docstring,
                CallableType.CONSTRUCTOR,
                prefer_docstring=prefer_docstring
            )
        except ValueError:
            constructor = None

        return ClassDescriptor(
            name,
            summary,
            description,
            constructor,
            attributes,
            properties,
            class_methods,
            methods,
            examples,
            module,
            package,
            file,
            bases
        )



"""Raises"""

from .common import Descriptor


class RaisesDescriptor(Descriptor):
    """A raises descriptor"""

    def __init__(
            self,
            type_: str,
            description: str
    ) -> None:
        """A raises descriptor

        Args:
            type_ (str): The type of exception raised
            description (str): The exception description
        """
        self.type = type_
        self.description = description

    @property
    def descriptor_type(self) -> str:
        return "raises"

    def __repr__(self) -> str:
        return f'{self.type} - {self.description}'



"""Utilities"""

import importlib
from inspect import Parameter
import re
from typing import Any, Optional, Union
import xml.etree.ElementTree as etree

from docstring_parser import Docstring, DocstringParam, DocstringReturns


def import_from_string(import_str: str) -> Any:
    """Import some python object from a given string

    Args:
        import_str (str): The import string

    Raises:
        ImportError: If the module could not be imported.
        ValueError: If the module could not be imported.

    Returns:
        Any: The imported object
    """
    module_str, _, attr_str = import_str.partition(":")

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        module_name = module_str.split(".", 1)[0]
        if exc.name != module_name:
            raise exc from None
        raise ValueError(f"Could not import module {module_str!r}.") from exc

    if not attr_str:
        return module

    try:
        return getattr(module, attr_str)
    except AttributeError as exc:
        raise ValueError(
            f"Attribute {attr_str!r} not found in module {module_str!r}."
        ) from exc


def add_tag(tag: str, class_name: Optional[str], parent: etree.Element) -> etree.Element:
    element = etree.SubElement(parent, tag)
    if class_name:
        element.set('class', class_name)
    return element


def add_text_tag(tag: str, text: str, klass: Optional[str], parent: etree.Element) -> etree.Element:
    element = add_tag(tag, klass, parent)
    element.text = text
    return element


def add_span_tag(text: str, klass: Optional[str], parent: etree.Element) -> etree.Element:
    return add_text_tag('span', text, klass, parent)


def find_docstring_param(name: str, docstring: Docstring) -> Optional[DocstringParam]:
    return next(
        (
            param
            for param in docstring.params
            if param.arg_name == name
        ),
        None
    ) if docstring else None


def get_type_name(
        annotation: Any,
        docstring_param: Optional[Union[DocstringParam, DocstringReturns]]
) -> str:
    """Get the type name

    Args:
        annotation (Any): The type annotation
        docstring_param (Optional[Union[DocstringParam, DocstringReturns]]):
            The docstring param

    Returns:
        str: The type description
    """
    type_name = docstring_param.type_name if docstring_param else None
    if type_name:
        return type_name
    if annotation is Parameter.empty:
        return 'Any'

    type_name = getattr(annotation, '__name__', None)
    if not type_name:
        type_name = getattr(annotation, '_', None)
    if not type_name:
        type_name = str(annotation)

    return re.sub(
        r'[^a-zA-Z_]*typing\.([a-zA-Z_][a-zA-Z_0-9]*)',
        r'\1',
        type_name
    )
